import EventEmitter from 'eventemitter3';
import { decodeDate, keyRemap, union, difference } from './utils';
import IMClient from './im-client';
import {
  GenericCommand,
  ConvCommand,
  JsonObjectMessage,
  DirectCommand,
  LogsCommand,
  ReadCommand,
  ReadTuple,
} from '../proto/message';
import { run as runSignatureFactory } from './signature-factory-runner';
import { createError } from './errors';
import Message from './messages/message';
import isEmpty from 'lodash/isEmpty';
import isPlainObject from 'lodash/isPlainObject';
import { default as d } from 'debug';

const debug = d('LC:Conversation');

export default class Conversation extends EventEmitter {
  /**
   * 无法直接实例化，请使用 {@link IMClient#createConversation} 创建新的对话
   */
  constructor(data, client) {
    super();
    Object.assign(this, {
      /**
       * @var id {String} 对话 id，对应 _Conversation 表中的 objectId
       * @memberof Conversation#
       */
      // id,
      /**
       * @var creator {String} 对话创建者
       * @memberof Conversation#
       */
      // creator,
      /**
       * @var createdAt {Date} 对话创建时间
       * @memberof Conversation#
       */
      // createdAt,
      /**
       * @var updatedAt {Date} 对话更新时间
       * @memberof Conversation#
       */
      // updatedAt,
      /**
       * @var lastMessageAt {Date} 最后一条消息时间
       * @memberof Conversation#
       */
      // lastMessageAt,
      /**
       * @var lastMessage {?Message} 最后一条消息
       * @memberof Conversation#
       */
      // lastMessage,
      /**
       * 对该对话设置了静音的用户列表
       * @memberof Conversation#
       * @type {?String[]}
       */
      mutedMembers: [],
      /**
       * 参与该对话的用户列表
       * @memberof Conversation#
       * @type {String[]}
       */
      members: [],
      _attributes: {},
      /**
       * 暂态对话标记
       * @memberof Conversation#
       * @type {Boolean}
       */
      transient: false,
      /**
       * 当前用户静音该对话标记
       * @memberof Conversation#
       * @type {Boolean}
       */
      muted: false,
      /**
       * 当前用户在该对话的未读消息数
       * @memberof Conversation#
       * @type {Number}
       */
      unreadMessagesCount: 0,
    }, keyRemap({
      attributes: '_attributes',
      name: '_name',
    }, data));
    this.members = Array.from(new Set(this.members));
    if (client instanceof IMClient) {
      this._client = client;
    } else {
      throw new TypeError('Conversation must be initialized with a client');
    }
    [
      'kicked',
      'membersjoined',
      'membersleft',
      'message',
    ].forEach(event => this.on(
      event,
      payload => this._debug(`${event} event emitted.`, payload)
    ));
  }

  set createdAt(value) {
    this._createdAt = decodeDate(value);
  }
  get createdAt() {
    return this._createdAt;
  }
  set updatedAt(value) {
    this._updatedAt = decodeDate(value);
  }
  get updatedAt() {
    return this._updatedAt;
  }
  set lastMessageAt(value) {
    this._lastMessageAt = decodeDate(value);
  }
  get lastMessageAt() {
    return this._lastMessageAt;
  }

  /**
   * 对话额外属性
   * @type {Object}
   */
  get attributes() {
    if (typeof this._pendingAttributes !== 'undefined') {
      return this._pendingAttributes;
    }
    return this._attributes;
  }
  set attributes(value) {
    this.setAttributes(value);
  }
  /**
   * 设置对话额外属性
   * @param {Object} map    key-value 对
   * @param {assign} [assign=false] 使用 Object.assign 更新属性，而不是替换整个 attributes
   * @return {Conversation} self
   */
  setAttributes(map, assign = false) {
    this._debug(`set attributes: value=${JSON.stringify(map)}, assign=${assign}`);
    if (!isPlainObject(map)) {
      throw new TypeError('attributes must be a plain object');
    }
    if (!assign) {
      this._pendingAttributes = map;
    } else {
      this._pendingAttributes = Object.assign({}, this.attributes, map);
    }
    return this;
  }
  /**
   * 设置对话额外属性
   * @param {String} key
   * @param {*} value
   * @return {Conversation} self
   */
  setAttribute(key, value) {
    if (typeof this._pendingAttributes === 'undefined') {
      this._pendingAttributes = {};
    }
    this._pendingAttributes[key] = value;
    return this;
  }
  /**
   * 对话名字
   * @type {String}
   */
  get name() {
    if (typeof this._pendingName !== 'undefined') {
      return this._pendingName;
    }
    return this._name;
  }
  set name(value) {
    this.setName(value);
  }
  /**
   * 设置对话名字
   * @param {String} value
   * @return {Conversation} self
   */
  setName(value) {
    this._debug(`set name: ${value}`);
    this._pendingName = value;
    return this;
  }

  _debug(...params) {
    debug(...params, `[${this.id}]`);
  }

  _send(command, ...args) {
    /* eslint-disable no-param-reassign */
    if (command.cmd === null) {
      command.cmd = 'conv';
    }
    if (command.cmd === 'conv' && command.convMessage === null) {
      command.convMessage = new ConvCommand();
    }
    if (command.convMessage && command.convMessage.cid === null) {
      command.convMessage.cid = this.id;
    }
    /* eslint-enable no-param-reassign */
    return this._client._send(command, ...args);
  }
  /**
   * 保存当前对话的属性至服务器
   * @return {Promise.<Conversation>} self
   */
  save() {
    this._debug('save');
    const attr = {};
    if (typeof this._pendingAttributes !== 'undefined') {
      attr.attr = this._pendingAttributes;
    }
    if (typeof this._pendingName !== 'undefined') {
      attr.name = this._pendingName;
    }
    if (isEmpty(attr)) {
      this._debug('nothing touched, resolve with self');
      return Promise.resolve(this);
    }
    this._debug(`attr: ${JSON.stringify(attr)}`);
    const convMessage = new ConvCommand({
      attr: new JsonObjectMessage({
        data: JSON.stringify(attr),
      }),
    });
    return this
      ._send(new GenericCommand({
        op: 'update',
        convMessage,
      }))
      .then(resCommand => {
        this.updatedAt = resCommand.convMessage.udate;
        if (typeof this._pendingAttributes !== 'undefined') {
          this._attributes = this._pendingAttributes;
          delete this._pendingAttributes;
        }
        if (typeof this._pendingName !== 'undefined') {
          this._name = this._pendingName;
          delete this._pendingName;
        }
        return this;
      });
  }

  /**
   * 从服务器更新对话的属性
   * @return {Promise.<Conversation>} self
   */
  fetch() {
    return this
      ._client
      .getQuery()
      .equalTo('objectId', this.id)
      .find()
      .then(() => this);
  }

  /**
   * 静音，客户端拒绝收到服务器端的离线推送通知
   * @return {Promise.<Conversation>} self
   */
  mute() {
    this._debug('mute');
    return this._send(new GenericCommand({
      op: 'mute',
    })).then(() => {
      if (!this.transient) {
        this.muted = true;
        this.mutedMembers = union(this.mutedMembers, [this._client.id]);
      }
      return this;
    });
  }

  /**
   * 取消静音
   * @return {Promise.<Conversation>} self
   */
  unmute() {
    this._debug('unmute');
    return this._send(new GenericCommand({
      op: 'unmute',
    })).then(() => {
      if (!this.transient) {
        this.muted = false;
        this.mutedMembers = difference(this.mutedMembers, [this._client.id]);
      }
      return this;
    });
  }

  /**
   * 获取对话人数，或暂态对话的在线人数
   * @return {Promise.<Number>}
   */
  count() {
    this._debug('unmute');
    return this._send(new GenericCommand({
      op: 'count',
    })).then(resCommand => resCommand.convMessage.count);
  }

  /**
   * 增加成员
   * @param {String|String[]} clientIds 新增成员 client id
   * @return {Promise.<Conversation>} self
   */
  add(clientIds) {
    this._debug('add', clientIds);
    if (typeof clientIds === 'string') {
      clientIds = [clientIds]; // eslint-disable-line no-param-reassign
    }
    const convMessage = new ConvCommand({
      m: clientIds,
    });
    return Promise.resolve(
      new GenericCommand({
        op: 'add',
        convMessage,
      })
    )
    .then(command => {
      if (this._client.options.conversationSignatureFactory) {
        const params = [this.id, this._client.id, clientIds.sort(), 'add'];
        return runSignatureFactory(this._client.options.conversationSignatureFactory, params)
          .then(signatureResult => {
            Object.assign(command.convMessage, keyRemap({
              signature: 's',
              timestamp: 't',
              nonce: 'n',
            }, signatureResult));
            return command;
          });
      }
      return command;
    })
    .then(this._send.bind(this))
    .then(() => {
      if (!this.transient) {
        this.members = union(this.members, clientIds);
      }
      return this;
    });
  }

  /**
   * 剔除成员
   * @param {String|String[]} clientIds 成员 client id
   * @return {Promise.<Conversation>} self
   */
  remove(clientIds) {
    this._debug('remove', clientIds);
    if (typeof clientIds === 'string') {
      clientIds = [clientIds]; // eslint-disable-line no-param-reassign
    }
    const convMessage = new ConvCommand({
      m: clientIds,
    });
    return Promise.resolve(
      new GenericCommand({
        op: 'remove',
        convMessage,
      })
    )
    .then(command => {
      if (this._client.options.conversationSignatureFactory) {
        const params = [this.id, this._client.id, clientIds.sort(), 'remove'];
        return runSignatureFactory(this._client.options.conversationSignatureFactory, params)
          .then(signatureResult => {
            Object.assign(command.convMessage, keyRemap({
              signature: 's',
              timestamp: 't',
              nonce: 'n',
            }, signatureResult));
            return command;
          });
      }
      return command;
    })
    .then(this._send.bind(this))
    .then(() => {
      if (!this.transient) {
        this.members = difference(this.members, clientIds);
      }
      return this;
    });
  }

  /**
   * （当前用户）加入该对话
   * @return {Promise.<Conversation>} self
   */
  join() {
    this._debug('join');
    return this.add(this._client.id);
  }

  /**
   * （当前用户）退出该对话
   * @return {Promise.<Conversation>} self
   */
  quit() {
    this._debug('quit');
    return this.remove(this._client.id);
  }

  /**
   * 发送消息
   * @param  {Message} message 消息，Message 及其子类的实例
   * @return {Promise.<Message>} 发送的消息
   */
  send(message) {
    this._debug(message, 'send');
    if (!(message instanceof Message)) {
      throw new TypeError(`${message} is not a Message`);
    }
    Object.assign(message, {
      cid: this.id,
      from: this._client.id,
    });
    let msg = message.toJSON();
    if (typeof msg !== 'string') {
      msg = JSON.stringify(msg);
    }
    let sendPromise = this._send(new GenericCommand({
      cmd: 'direct',
      directMessage: new DirectCommand({
        msg,
        cid: this.id,
        r: message.needReceipt,
        transient: message.transient,
        dt: message.id,
      }),
    }), !message.transient);
    if (!message.transient) {
      sendPromise = sendPromise.then(resCommand => {
        const {
          ackMessage: {
            uid,
            t,
            code,
            reason,
            appCode,
            },
          } = resCommand;
        if (code !== null) {
          throw createError({
            code, reason, appCode,
          });
        }
        Object.assign(message, {
          id: uid,
          timestamp: new Date(t.toNumber()),
        });
        this.lastMessage = message;
        this.lastMessageAt = message.timestamp;
        return message;
      });
    }
    return sendPromise;
  }

  /**
   * 查询消息记录
   * 如果仅需实现消息记录翻页查询需求，建议使用 {@link Conversation#createMessagesIterator}
   * @param  {Object} [options]
   * @param  {Date}   [options.beforeTime] 限制查询结果为小于这个该时间之前的消息，不传则为当前时间
   * @param  {String} [options.beforeMessageId] 限制查询结果为该消息之前的消息，需要与 beforeTime 同时使用，为防止某时刻有重复消息
   * @param  {Date}   [options.afterTime] 限制查询结果为大于这个该时间之前的消息
   * @param  {String} [options.afterMessageId] 限制查询结果为该消息之后的消息，需要与 afterTime 同时使用，为防止某时刻有重复消息
   * @param  {Number} [options.limit] 限制查询结果的数量，目前服务端默认为 20
   * @return {Promise.<Message[]>} 消息列表
   */
  queryMessages(options = {}) {
    this._debug('query messages', options);
    if (options.beforeMessageId && !options.beforeTime) {
      throw new Error('query option beforeMessageId must be used with option beforeTime');
    }
    if (options.afterMessageId && !options.afterTime) {
      throw new Error('query option afterMessageId must be used with option afterTime');
    }
    const conditions = keyRemap({
      beforeTime: 't',
      beforeMessageId: 'mid',
      afterTime: 'tt',
      afterMessageId: 'tmid',
      limit: 'l',
    }, options);
    if (conditions.t instanceof Date) {
      conditions.t = conditions.t.getTime();
    }
    if (conditions.tt instanceof Date) {
      conditions.tt = conditions.tt.getTime();
    }
    return this._send(new GenericCommand({
      cmd: 'logs',
      logsMessage: new LogsCommand(
        Object.assign(conditions, {
          cid: this.id,
        })
      ),
    })).then(resCommand =>
      resCommand.logsMessage.logs.map(log => {
        const messageProps = {
          id: log.msgId,
          cid: this.id,
          timestamp: new Date(log.timestamp.toNumber()),
          from: log.from,
        };
        const message = this._client._messageParser.parse(log.data);
        Object.assign(message, messageProps);
        return message;
      })
    );
  }

  /**
   * 获取消息翻页迭代器
   * @param  {Object} [options]
   * @param  {Date}   [options.beforeTime] 限制起始查询结果为小于这个该时间之前的消息，不传则为当前时间
   * @param  {String} [options.beforeMessageId] 限制起始查询结果为该消息之前的消息，需要与 beforeTime 同时使用，为防止某时刻有重复消息
   * @param  {Number} [options.limit] 限制每页查询结果的数量，目前服务端默认为 20
   * @return {AsyncIterater.<Promise.<IteratorResult<Message[]>>>} [AsyncIterator]{@link https://github.com/tc39/proposal-async-iteration}，调用其 next 方法返回获取下一页消息的 Promise
   * @example
   * var messageIterator = conversation.createMessagesIterator({ limit: 10 });
   * messageIterator.next().then(function(result) {
   *   // result: {
   *   //   value: [message1, ..., message10],
   *   //   done: false,
   *   // }
   * });
   * messageIterator.next().then(function(result) {
   *   // result: {
   *   //   value: [message11, ..., message20],
   *   //   done: false,
   *   // }
   * });
   * messageIterator.next().then(function(result) {
   *   // No more messages
   *   // result: { value: [], done: true }
   * });
   */
  createMessagesIterator({ beforeTime, beforeMessageId, limit } = {}) {
    let promise;
    return {
      next: () => {
        if (promise === undefined) {
          // first call
          promise = this.queryMessages({
            limit,
            beforeTime,
            beforeMessageId,
          });
        } else {
          promise = promise.then(prevMessages => {
            if (prevMessages.length === 0 || prevMessages.length < limit) {
              // no more messages
              return [];
            }
            return this.queryMessages({
              beforeTime: prevMessages[0].timestamp,
              beforeMessageId: prevMessages[0].id,
              limit,
            });
          });
        }
        return promise.then(value => ({
          value,
          done: value.length === 0 || value.length < limit,
        }));
      },
    };
  }

  /**
   * 将该会话标记为已读
   * @return {Promise.<Conversation>} self
   */
  markAsRead() {
    this._debug('mark as read');
    return this._send(new GenericCommand({
      cmd: 'read',
      readMessage: new ReadCommand({
        convs: [new ReadTuple({
          cid: this.id,
          timestamp: (this.lastMessageAt || new Date()).getTime(),
        })],
      }),
    }), false).then(() => {
      this.unreadMessagesCount = 0;
      return this;
    });
  }
}
